import contextlib
import os
import shutil
import tempfile
import subprocess

import lola
from expecter import expect

DEVNULL = open('/dev/null', 'w')


@contextlib.contextmanager
def temp_directory():
  tmpdir = tempfile.mkdtemp()
  try:
    yield tmpdir
  finally:
    shutil.rmtree(tmpdir)


def test_success_retcode():
  with lola.Runner() as runner:
    retcode = runner.call(['test_scripts/success.py'],
                          stdout=DEVNULL, stderr=DEVNULL)
  expect(retcode) == 0


def test_failure_retcode():
  with lola.Runner() as runner:
    retcode = runner.call(['test_scripts/fail.py'],
                          stdout=DEVNULL, stderr=DEVNULL)
  expect(retcode) == 1


def test_exception():
  with lola.Runner() as runner:
    retcode = runner.call(['test_scripts/exception.py'],
                          stdout=DEVNULL, stderr=DEVNULL)
  expect(retcode) == 255


def test_check_call_success():
  with lola.Runner() as runner:
    retcode = runner.check_call(['test_scripts/success.py'],
                                stdout=DEVNULL, stderr=DEVNULL)
  expect(retcode) == 0


def test_check_call_error():
  try:
    with lola.Runner() as runner:
      runner.check_call(['test_scripts/fail.py'],
                        stdout=DEVNULL, stderr=DEVNULL)
  except subprocess.CalledProcessError as e:
    expect(e.returncode) == 1
  else:
    assert False, 'Should have raise a CalledProcessError'


def test_output_to_file():
  with temp_directory() as tmpdir:
    outfile = os.path.join(tmpdir, 'out.txt')
    errfile = os.path.join(tmpdir, 'err.txt')

    with lola.Runner() as runner:
      runner.check_call(['test_scripts/output.py'],
                        stdout=open(outfile, 'w'), stderr=open(errfile, 'w'))

    expect(open(outfile).read()) == 'outputting\n'
    expect(open(errfile).read()) == 'erroring\n'


def test_stderr_to_stdout():
  with temp_directory() as tmpdir:
    outfile = os.path.join(tmpdir, 'out.txt')

    with lola.Runner() as runner:
      runner.check_call(['test_scripts/output.py'],
                        stdout=open(outfile, 'w'), stderr=subprocess.STDOUT)

    expect(open(outfile).read()) == 'outputting\nerroring\n'


def test_check_output():
  with lola.Runner() as runner:
    out = runner.check_output(['test_scripts/output.py'],
                              stderr=subprocess.STDOUT)
  expect(out) == 'outputting\nerroring\n'


def test_check_output_error():
  try:
    with lola.Runner() as runner:
      runner.check_output(['test_scripts/fail.py'], stderr=subprocess.STDOUT)
  except subprocess.CalledProcessError as e:
    expect(e.returncode) == 1
    expect(e.output) == 'failing'
  else:
    assert False, 'Should have raise a CalledProcessError'


def test_cwd():
  with temp_directory() as tmpdir:
    with lola.Runner() as runner:
      out = runner.check_output([os.path.abspath('test_scripts/pwd.py')],
                                cwd=tmpdir)
    # use samefile instead of string comparison since on OSX we get "/var" as
    # tmpdir, but "/private/var" from the pwd script
    assert os.path.samefile(out, tmpdir)


def test_env():
  env = os.environ.copy()
  env['LOLA_ENV_TEST'] = 'test value'

  with lola.Runner() as runner:
    out = runner.check_output(['test_scripts/env.py', 'LOLA_ENV_TEST'], env=env)

  expect(out) == 'test value'


def test_env_overrides_parent():
  env = os.environ.copy()
  # this key set in the parent process should not be inherited when
  # env is passed
  os.environ['LOLA_ENV_TEST_PARENT'] = 'set in parent'

  with lola.Runner() as runner:
    out = runner.check_output(['test_scripts/env.py', 'LOLA_ENV_TEST_PARENT'],
                              env=env)

  expect(out) == 'not set'


def test_env_inherit_from_parent():
  # when "env" is not set explicitly this should be passed through
  os.environ['LOLA_ENV_TEST_PARENT'] = 'set in parent'

  with lola.Runner() as runner:
    out = runner.check_output(['test_scripts/env.py', 'LOLA_ENV_TEST_PARENT'])

  expect(out) == 'set in parent'


def test_reclose_throws_error():
  runner = lola.Runner()
  runner.close()

  with expect.raises(ValueError):
    runner.close()


def test_call_with_closed_runner_throws_error():
  runner = lola.Runner()
  runner.close()

  with expect.raises(ValueError):
    runner.call(['test_scripts/success.py'], stdout=DEVNULL, stderr=DEVNULL)


def test_with_block_closes_runner():
  with lola.Runner() as runner:
    pass

  with expect.raises(ValueError):
    runner.close()


def test_with_block_closes_runner_after_error():
  with expect.raises(KeyError):
    with lola.Runner() as runner:
      raise KeyError()

  with expect.raises(ValueError):
    runner.close()



#!/usr/bin/python

import setuptools

setuptools.setup(
  name = 'lola',
  version = '0.1',
  license = 'BSD',
  description = 'Lola runs small Python scripts quickly',
  long_description = open('README.txt').read(),
  author = 'Matt Good',
  author_email = 'matt@matt-good.net',
  url = 'http://mgood.github.com/lola/',
  platforms = 'any',

  py_modules = ['lola'],

  zip_safe = True,
  verbose = False,
)



import os
import subprocess
import sys
import traceback
import _multiprocessing
from multiprocessing.connection import Listener, Client


class Runner(object):
  """Run a Python interpreter in the background for executing scripts.

  It implements a subprocess-like API, providing equivalent functions to
  ``call()``, ``check_call()``, and ``check_output()``.

  Example::

    with lola.Runner() as runner:
      retcode = runner.call(['myscript.py', 'arg1', 'arg2'])
      runner.check_call(['myscript.py', 'arg3', 'arg4'])
      out = runner.check_output(['myscript.py', 'arg3', 'arg4'])

      # redirect stdout, stderr, and provide a custom environment dict
      runner.call(['myscript.py', 'arg1', 'arg2'],
                  stdout=open('out.txt', 'w'),
                  stderr=subprocess.STDOUT,
                  env=env)

  """
  def __init__(self, python=None):
    if python is None:
      python = sys.executable

    self._listener = listener = Listener()
    subprocess.Popen([python, __file__, listener.address])
    self._cnx = listener.accept()

  def close(self):
    self._expect_connected()
    try:
      self._cnx.close()
    finally:
      try:
        self._listener.close()
      finally:
        self._cnx = self._listener = None

  def __enter__(self):
    return self

  def __exit__(self, *args):
    self.close()

  def call(self, *popenargs, **kwargs):
    """Run Python script with arguments.  Wait for command to complete, then
    return the return code.

    The arguments are the same as for the subprocess.Popen constructor.
    """
    retcode, _, _ = self._call(*popenargs, **kwargs)
    return retcode

  def check_call(self, *popenargs, **kwargs):
    """Run Python script with arguments.  Wait for command to complete.  If
    the exit code was zero then return, otherwise raise
    subprocess.CalledProcessError.  The CalledProcessError object will have the
    return code in the returncode attribute.

    The arguments are the same as for the Popen constructor.
    """
    retcode = self.call(*popenargs, **kwargs)
    if retcode:
      cmd = kwargs.get("args")
      if cmd is None:
        cmd = popenargs[0]
      raise subprocess.CalledProcessError(retcode, cmd)
    return retcode

  def check_output(self, *popenargs, **kwargs):
    """Run command with arguments and return its output as a byte string.

    If the exit code was non-zero it raises a CalledProcessError.  The
    CalledProcessError object will have the return code in the returncode
    attribute and output in the output attribute.

    The arguments are the same as for the Popen constructor.

    The stdout argument is not allowed as it is used internally.
    To capture standard error in the result, use stderr=STDOUT.
    """
    retcode, stdout, _ = self._call(stdout=subprocess.PIPE,
                                    *popenargs, **kwargs)
    if retcode:
      cmd = kwargs.get("args")
      if cmd is None:
        cmd = popenargs[0]
      err = subprocess.CalledProcessError(retcode, cmd)
      err.output = stdout
      raise err
    return stdout

  def _call(self, args, bufsize=0, executable=None,
            stdin=None, stdout=None, stderr=None,
            preexec_fn=None, close_fds=False, shell=False,
            cwd=None, env=None, universal_newlines=False,
            startupinfo=None, creationflags=0):
    self._expect_connected()

    fds = []
    def mkfd(fp):
      if fp in (None, subprocess.STDOUT, subprocess.PIPE):
        return fp
      elif isinstance(fp, int):
        fds.append(fp)
      else:
        fds.append(fp.fileno())
      return len(fds) - 1

    stdin = mkfd(stdin)
    stdout = mkfd(stdout)
    stderr = mkfd(stderr)

    self._cnx.send(((args, bufsize, executable, stdin, stdout, stderr,
                     preexec_fn, close_fds, shell, cwd, env, universal_newlines,
                     startupinfo, creationflags),
                    len(fds)))
    for fd in fds:
      _multiprocessing.sendfd(self._cnx.fileno(), fd)
    return self._cnx.recv()

  def _expect_connected(self):
    if self._cnx is None:
      raise ValueError('Runner is closed')


def listen(address):
  cnx = Client(address)
  while True:
    try:
      args, num_fds = cnx.recv()
    except EOFError:
      break
    fds = dict(
      (idx, _multiprocessing.recvfd(cnx.fileno()))
      for idx in range(num_fds)
    )
    proc = _PyPopen(*args, _fd_map=fds)
    stdout, stderr = proc.communicate()
    retcode = proc.poll()
    cnx.send((retcode, stdout, stderr))


class _PyPopen(subprocess.Popen):
  def __init__(self, args, bufsize=0, executable=None,
               stdin=None, stdout=None, stderr=None,
               preexec_fn=None, close_fds=False, shell=False,
               cwd=None, env=None, universal_newlines=False,
               startupinfo=None, creationflags=0, _fd_map=None):
    """
      Run a Python script in a forked interpreter.

      This class uses the same interface as Popen, and can be used as a
      replacement to run Python scripts in a forked process, without executing
      a new Python interpreter.
    """
    if shell:
      raise ValueError()

    if _fd_map is not None:
      stdin = _fd_map.get(stdin, stdin)
      stdout = _fd_map.get(stdout, stdout)
      stderr = _fd_map.get(stderr, stderr)

    def exec_python(args=args, executable=executable):
      if isinstance(args, basestring):
        args = [args]
      else:
        args = list(args)

      if executable is None:
        executable = args[0]

      if preexec_fn:
        preexec_fn()

      sys.argv = args

      if env is not None:
        for key in list(os.environ):
          if key not in env:
            del os.environ[key]
        os.environ.update(env)

      try:
        p_globals = dict((k,getattr(__builtins__, k))
                         for k in dir(__builtins__))
        p_globals['__name__'] = '__main__'
        p_globals['__file__'] = os.path.abspath(executable)
        execfile(executable, p_globals)
      except SystemExit as e:
        sys.stdout.flush()
        if isinstance(e.code, int):
          os._exit(e.code)
        elif e.code is None:
          os._exit(0)
        else:
          print >>sys.stderr, e.code
          os._exit(1)
      except:
        sys.stdout.flush()
        traceback.print_exc()
        os._exit(255)
      else:
        sys.stdout.flush()
        os._exit(0)

    subprocess.Popen.__init__(self, args, bufsize, executable, stdin, stdout,
                              stderr, exec_python, close_fds, shell, cwd, env,
                              universal_newlines, startupinfo, creationflags)


if __name__ == '__main__':
  address = sys.argv[1]
  listen(address)
